// Gaia Sky port of shadertoy shader "Helix Nebula" by izutionix.
// Licensed under CC-BY-NC-SA.
// Source: https://www.shadertoy.com/view/cdK3Wy
#version 330 core

// Uniforms which are always available
uniform mat4 u_worldTrans;
uniform vec3 u_cameraPos;
uniform float u_bodySize;
uniform vec2 u_cameraNearFar;
uniform float u_cameraK;
uniform float u_time;
uniform float u_simuTime;
// Unit conversion.
uniform float u_kmToU;

#ifdef ssrFlag
#include <shader/lib/ssr.frag.glsl>
#endif // ssrFlag

// INPUT
struct VertexData {
    vec2 texCoords;
    vec3 normal;
    vec3 viewDir;
    vec3 ambientLight;
    float opacity;
    vec4 color;
    vec3 fragPosWorld;
    mat3 tbn;
};
in VertexData v_data;

// OUTPUT
layout(location = 0) out vec4 fragColor;
layout(location = 1) out vec4 layerBuffer;

#include <shader/lib/logdepthbuff.glsl>
#include <$data/catalog-nebulae/shader/lib/nebulae-lib.glsl>

#define DIST_SCALE (u_kmToU * 1.0e6)
#define SCALE_FACTOR 4.5e-5
#define FACTOR (1.0 / DIST_SCALE * SCALE_FACTOR)

#define ITERATIONS 80.0

float Ring(vec3 p) {
    p.x *= 1.3;
    p.y *= 0.9;
    vec2 q = vec2(length(p.xy) - 2.2, p.z);
    return length(q);
}
float Disk(vec3 p, vec3 t) {
    vec2 q = vec2(length2(p.xy) - t.x, p.z);
    return max(length8(q) - t.y, abs(p.z) - t.z);
}
float Sphere(vec3 p, float s) {
    return length(p) - s;
}

const float nudge = 0.9; // size of perpendicular vector
float normalizer = 1.0 / sqrt(1.0 + nudge * nudge); // pythagorean theorem on that perpendicular to maintain scale
float SpiralNoiseC(vec3 p) {
    float n = 0.0; // noise amount
    float iter = 2.0;
    for (int i = 0; i < 8; i++) {
        // add sin and cos scaled inverse with the frequency
        n += -abs(sin(p.y * iter) + cos(p.x * iter)) / iter; // abs for a ridged look
        // rotate by adding perpendicular and scaling down
        p.xy += vec2(p.y, -p.x) * nudge;
        p.xy *= normalizer;
        // rotate on other axis
        p.xz += vec2(p.z, -p.x) * nudge;
        p.xz *= normalizer;
        // increase the frequency
        iter *= 1.733733;
    }
    return n;
}

float NebulaNoise(vec3 p) {
    float final = Sphere(p.xzy, 3.5);
    final += fbm(p * 10.0);
    final += SpiralNoiseC(p.zxy * 0.415);

    return final * 2.5;
}

float smin(float a, float b, float k) {
    float h = clamp(0.5 + 0.5 * (b - a) / k, 0.0, 1.0);
    return mix(b, a, h) - k * h * (1.0 - h);
}
float Xor(float d1, float d2) {
    return max(min(d1, d2), -max(d1, d2));
}

float map(vec3 p) {
    //p = vec3(-p.x, -p.z, -p.y);
    vec4 q = get_quat_rotation(vec3(0.0, 0.0, 1.0), 60.0 * PI / 180.0);
    p = rotate(p, q);
    vec4 q1 = get_quat_rotation(vec3(0.0, 1.0, 0.0), 90.0 * PI / 180.0);
    p = rotate(p, q1);

    float d1 = Disk(p, vec3(2.2, 1.0, 0.3)) + noise((p + 0.1) * 17.0) * 0.8;
    float d2 = Ring(p);
    float d3 = abs(NebulaNoise(p) * 0.8) + 0.07;
    return Xor(d3, smin(d1, d2, 1.0));
}
//--------------------------------------------------------------

vec3 computeColor(float density, float radius) {
    vec3 result = mix(vec3(1.0, 0.9, 0.8), vec3(0.4, 0.15, 0.1), density);
    vec3 colCenter = 7.0 * vec3(0.5, 0.99, 1.0);
    vec3 colEdge = 2.0 * vec3(0.7, 0.67, 0.5);
    vec3 colOuter = vec3(1.0, 0.4, 0.42);
    result *= mix(mix(colCenter, colEdge, min((radius) * 0.9, 1.15)), colOuter, radius * 0.19);

    return result;
}

struct RaymarchResult {
    vec4 color;
    float depth;
};

RaymarchResult raymarch(vec3 camPos, vec3 rayDir, vec3 objPos, vec2 dists) {
    // Ray direction.
    vec3 rd = rayDir;
    // Ray origin.
    vec3 ro = camPos;

    // ld, td: local, total density
    // w: weighting factor
    float ld = 0.0, td = 0.0, w = 0.0;

    // d: distance function
    float d = 1.0;
    // depth: depth from the camera
    float depth = -1.0;

    const float h = 0.1;

    vec4 sum = vec4(0.0);

    float min_dist = dists.x, max_dist = dists.y;

    if (min_dist >= 0.0 && max_dist >= 0.0) {
        // t: length of the ray
        float t = min_dist;

        // raymarch loop
        for (int i = 0; i < ITERATIONS; i++) {
            vec3 pos = ro + t * rd;

            // Loop break conditions.
            if (td > 0.9 || d < 0.02 * t || sum.a > 0.99 || t > max_dist) break;

            // evaluate distance function
            float d = map(pos);

            // change this string to control density
            d = max(d, 0.0);

            // point light calculations
            vec3 ldst = vec3(0.0) - pos;
            float lDist = max(length(ldst), 0.001);

            // the color of light
            float _T = lDist * 2.2 + 2.0;
            vec3 lightColor = 0.4 + 0.45 * cos(_T + PI * 0.5 * vec3(-0.6, 0.25, 1.4));

            // Bloom
            sum.rgb += (lightColor / exp(lDist * lDist * lDist * 0.03) / 35.0);

            if (d < h) {
                // compute local density
                ld = h - d;

                // compute weighting factor
                w = (1.0 - td) * ld;

                // accumulate density
                td += w + 1.0 / 40.0;

                vec4 col = vec4(computeColor(td, lDist), td);

                // emission
                sum += sum.a * vec4(sum.rgb, 0.0) * 0.2;

                // uniform scale density
                col.a *= 0.2;
                // colour by alpha
                col.rgb *= col.a;
                // alpha blend in contribution
                sum = sum + col * (1.0 - sum.a);

                if (depth < 0.0) {
                    depth = t;
                }
            }
            td += 1.0 / 70.0;
            // Optimize step size near the camera
            t += max(d * 0.1 * max(min(length(ldst), length(ro)), 1.0), 0.01);
        }
        td *= smoothstep(0.0, 0.8, td);

        // simple scattering
        sum *= 1.0 / exp(ld * 0.2) * 0.4;
        sum = clamp(sum, 0.0, 1.0);
        sum.xyz = sum.xyz * sum.xyz * (3.0 - 2.0 * sum.xyz);
    }
    RaymarchResult result;
    result.color = vec4(sum.xyz, td);
    result.depth = depth;
    return result;
}

void main() {
    // Directions and positions.
    // Object is alwyas at origin in our reference frame.
    vec3 camPos = u_cameraPos * FACTOR;
    vec3 rayDir = normalize(v_data.fragPosWorld);
    vec3 objPos = vec3(0.0, 0.0, 0.0);
    float radius = u_bodySize * SCALE_FACTOR / DIST_SCALE;

    vec2 dists = sphereIntersectionDistances(camPos, rayDir, objPos, radius);

    // Ray marching.
    RaymarchResult result = raymarch(camPos, rayDir, objPos, dists);
    fragColor = result.color;

    // Render outline
    // vec2 uv = v_data.texCoords;
    // if (uv.x < 0.005 || uv.x > 0.995 || uv.y < 0.005 || uv.y > 0.995) {
    //     d = length(v_data.fragPosWorld);
    //     fragColor = vec4(1.0, 0.0, 0.0, 1.0);
    // }

    // Logarithmic depth buffer.
    if (result.depth > 0.0) {
        gl_FragDepth = getDepthValue(result.depth * DIST_SCALE / SCALE_FACTOR, u_cameraNearFar.y, u_cameraK);
    } else {
        gl_FragDepth = getDepthValue(1.0e30, u_cameraNearFar.y, u_cameraK);
    }

    #ifdef ssrFlag
    ssrBuffers();
    #endif // ssrFlag
}
